"""Heracls - Slayer of Hydra"""

__version__ = "0.5.1"

from . import cli, core, patch  # noqa: F401
from .cli import ArgumentParser, choice  # noqa: F401
from .core import (  # noqa: F401
    from_dict,
    from_dotlist,
    from_omega,
    from_yaml,
    to_dict,
    to_omega,
    to_yaml,
)
